@@ -1,7 +1,6 @@
 # pylint:  disable=too-many-lines
 # (Yes, it has a point!)
 
-from __future__ import division, absolute_import, print_function
 
 __copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
 
@@ -26,21 +25,27 @@ THE SOFTWARE.
 """
 
 
-from functools import reduce
+import re
+from functools import reduce, wraps
 import operator
 import sys
 import logging
 from typing import (
         Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, TypeVar)
+import builtins
 
 
-import six
-from six.moves import range, zip, intern, input
+from sys import intern
 
 
 decorator_module = __import__("decorator", level=0)
 my_decorator = decorator_module.decorator
 
+# These are deprecated and will go away in 2022.
+all = builtins.all
+any = builtins.any
+
+
 __doc__ = """
 A Collection of Utilities
 =========================
@@ -120,6 +125,11 @@ Name generation
 .. autofunction:: generate_numbered_unique_names
 .. autoclass:: UniqueNameGenerator
 
+Deprecation Warnings
+--------------------
+
+.. autofunction:: deprecate_keyword
+
 Functions for dealing with (large) auxiliary files
 --------------------------------------------------
 
@@ -153,12 +163,96 @@ Sorting in natural order
 
 .. autofunction:: natorder
 .. autofunction:: natsorted
+
+Backports of newer Python functionality
+---------------------------------------
+
+.. autofunction:: resolve_name
+
+Hashing
+-------
+
+.. autofunction:: unordered_hash
+
+Type Variables Used
+-------------------
+
+.. class:: T
+
+    Any type.
+
+.. class:: F
+
+    Any callable.
 """
 
 # {{{ type variables
 
 T = TypeVar("T")
-F = TypeVar('F', bound=Callable[..., Any])
+F = TypeVar("F", bound=Callable[..., Any])
+
+# }}}
+
+
+# {{{ code maintenance
+
+class MovedFunctionDeprecationWrapper:
+    def __init__(self, f, deadline=None):
+        if deadline is None:
+            deadline = "the future"
+
+        self.f = f
+        self.deadline = deadline
+
+    def __call__(self, *args, **kwargs):
+        from warnings import warn
+        warn(f"This function is deprecated and will go away in {self.deadline}. "
+            f"Use {self.f.__module__}.{self.f.__name__} instead.",
+            DeprecationWarning, stacklevel=2)
+
+        return self.f(*args, **kwargs)
+
+
+def deprecate_keyword(oldkey: str,
+        newkey: Optional[str] = None, *,
+        deadline: Optional[str] = None):
+    """Decorator used to deprecate function keyword arguments.
+
+    :arg oldkey: deprecated argument name.
+    :arg newkey: new argument name that serves the same purpose, if any.
+    :arg deadline: expected time frame for the removal of the deprecated argument.
+    """
+    from warnings import warn
+
+    if deadline is None:
+        deadline = "the future"
+
+    def wrapper(func):
+        @wraps(func)
+        def inner_wrapper(*args, **kwargs):
+            if oldkey in kwargs:
+                if newkey is None:
+                    warn(f"The '{oldkey}' keyword is deprecated and will "
+                            f"go away in {deadline}.",
+                            DeprecationWarning, stacklevel=2)
+                else:
+                    warn(f"The '{oldkey}' keyword is deprecated and will "
+                            f"go away in {deadline}. "
+                            f"Use '{newkey}' instead.",
+                            DeprecationWarning, stacklevel=2)
+
+                    if newkey in kwargs:
+                        raise ValueError(f"Cannot use '{oldkey}' "
+                                f"and '{newkey}' in the same call.")
+
+                    kwargs[newkey] = kwargs[oldkey]
+                    del kwargs[oldkey]
+
+            return func(*args, **kwargs)
+
+        return inner_wrapper
+
+    return wrapper
 
 # }}}
 
@@ -227,7 +321,7 @@ def norm_p(iterable, p):
     return sum(i**p for i in iterable)**(1/p)
 
 
-class Norm(object):
+class Norm:
     def __init__(self, p):
         self.p = p
 
@@ -241,7 +335,7 @@ class Norm(object):
 
 # {{{ record
 
-class RecordWithoutPickling(object):
+class RecordWithoutPickling:
     """An aggregate of named sub-variables. Assumes that each record sub-type
     will be individually derived from this class.
     """
@@ -262,7 +356,7 @@ class RecordWithoutPickling(object):
         if valuedict is not None:
             kwargs.update(valuedict)
 
-        for key, value in six.iteritems(kwargs):
+        for key, value in kwargs.items():
             if key not in exclude:
                 fields.add(key)
                 setattr(self, key, value)
@@ -280,9 +374,9 @@ class RecordWithoutPickling(object):
         return self.__class__(**self.get_copy_kwargs(**kwargs))
 
     def __repr__(self):
-        return "%s(%s)" % (
+        return "{}({})".format(
                 self.__class__.__name__,
-                ", ".join("%s=%r" % (fld, getattr(self, fld))
+                ", ".join("{}={!r}".format(fld, getattr(self, fld))
                     for fld in self.__class__.fields
                     if hasattr(self, fld)))
 
@@ -298,7 +392,7 @@ class RecordWithoutPickling(object):
         # This method is implemented to avoid pylint 'no-member' errors for
         # attribute access.
         raise AttributeError(
-                "'%s' object has no attribute '%s'" % (
+                "'{}' object has no attribute '{}'".format(
                     self.__class__.__name__, name))
 
 
@@ -306,10 +400,10 @@ class Record(RecordWithoutPickling):
     __slots__: List[str] = []
 
     def __getstate__(self):
-        return dict(
-                (key, getattr(self, key))
+        return {
+                key: getattr(self, key)
                 for key in self.__class__.fields
-                if hasattr(self, key))
+                if hasattr(self, key)}
 
     def __setstate__(self, valuedict):
         try:
@@ -317,7 +411,7 @@ class Record(RecordWithoutPickling):
         except AttributeError:
             self.__class__.fields = fields = set()
 
-        for key, value in six.iteritems(valuedict):
+        for key, value in valuedict.items():
             fields.add(key)
             setattr(self, key, value)
 
@@ -330,7 +424,7 @@ class Record(RecordWithoutPickling):
 
 
 class ImmutableRecordWithoutPickling(RecordWithoutPickling):
-    "Hashable record. Does not explicitly enforce immutability."
+    """Hashable record. Does not explicitly enforce immutability."""
     def __init__(self, *args, **kwargs):
         RecordWithoutPickling.__init__(self, *args, **kwargs)
         self._cached_hash = None
@@ -350,7 +444,7 @@ class ImmutableRecord(ImmutableRecordWithoutPickling, Record):
 # }}}
 
 
-class Reference(object):
+class Reference:
     def __init__(self, value):
         self.value = value
 
@@ -363,7 +457,7 @@ class Reference(object):
         self.value = value
 
 
-class FakeList(object):
+class FakeList:
     def __init__(self, f, length):
         self._Length = length
         self._Function = f
@@ -381,7 +475,7 @@ class FakeList(object):
 
 # {{{ dependent dictionary ----------------------------------------------------
 
-class DependentDictionary(object):
+class DependentDictionary:
     def __init__(self, f, start=None):
         if start is None:
             start = {}
@@ -412,13 +506,13 @@ class DependentDictionary(object):
         return list(self._Dictionary.keys())
 
     def iteritems(self):
-        return six.iteritems(self._Dictionary)
+        return self._Dictionary.items()
 
     def iterkeys(self):
-        return six.iterkeys(self._Dictionary)
+        return self._Dictionary.keys()
 
     def itervalues(self):
-        return six.itervalues(self._Dictionary)
+        return self._Dictionary.values()
 
 # }}}
 
@@ -512,13 +606,13 @@ def memoize(*args: F, **kwargs: Any) -> F:
         which computes and returns the cache key.
     """
 
-    use_kw = bool(kwargs.pop('use_kwargs', False))
+    use_kw = bool(kwargs.pop("use_kwargs", False))
 
     default_key_func: Optional[Callable[..., Any]]
 
     if use_kw:
         def default_key_func(*inner_args, **inner_kwargs):  # noqa pylint:disable=function-redefined
-            return inner_args, frozenset(six.iteritems(inner_kwargs))
+            return inner_args, frozenset(inner_kwargs.items())
     else:
         default_key_func = None
 
@@ -574,7 +668,7 @@ def memoize(*args: F, **kwargs: Any) -> F:
 FunctionValueCache = memoize
 
 
-class _HasKwargs(object):
+class _HasKwargs:
     pass
 
 
@@ -586,12 +680,13 @@ def memoize_on_first_arg(function, cache_dict_name=None):
     """
 
     if cache_dict_name is None:
-        cache_dict_name = intern("_memoize_dic_"
-                + function.__module__ + function.__name__)
+        cache_dict_name = intern(
+                f"_memoize_dic_{function.__module__}{function.__name__}"
+                )
 
     def wrapper(obj, *args, **kwargs):
         if kwargs:
-            key = (_HasKwargs, frozenset(six.iteritems(kwargs))) + args
+            key = (_HasKwargs, frozenset(kwargs.items())) + args
         else:
             key = args
 
@@ -599,7 +694,7 @@ def memoize_on_first_arg(function, cache_dict_name=None):
             return getattr(obj, cache_dict_name)[key]
         except AttributeError:
             result = function(obj, *args, **kwargs)
-            setattr(obj, cache_dict_name, {key: result})
+            object.__setattr__(obj, cache_dict_name, {key: result})
             return result
         except KeyError:
             result = function(obj, *args, **kwargs)
@@ -607,7 +702,7 @@ def memoize_on_first_arg(function, cache_dict_name=None):
             return result
 
     def clear_cache(obj):
-        delattr(obj, cache_dict_name)
+        object.__delattr__(obj, cache_dict_name)
 
     from functools import update_wrapper
     new_wrapper = update_wrapper(wrapper, function)
@@ -618,9 +713,15 @@ def memoize_on_first_arg(function, cache_dict_name=None):
 
 def memoize_method(method: F) -> F:
     """Supports cache deletion via ``method_name.clear_cache(self)``.
+
+    .. versionchanged:: 2021.2
+
+        Can memoize methods on classes that do not allow setting attributes
+        (e.g. by overwritting ``__setattr__``).
     """
 
-    return memoize_on_first_arg(method, intern("_memoize_dic_"+method.__name__))
+    return memoize_on_first_arg(method,
+            intern(f"_memoize_dic_{method.__name__}"))
 
 
 class keyed_memoize_on_first_arg:  # noqa: N801
@@ -640,8 +741,7 @@ class keyed_memoize_on_first_arg:  # noqa: N801
         self.cache_dict_name = cache_dict_name
 
     def _default_cache_dict_name(self, function):
-        return intern("_memoize_dic_"
-                + function.__module__ + function.__name__)
+        return intern(f"_memoize_dic_{function.__module__}{function.__name__}")
 
     def __call__(self, function):
         cache_dict_name = self.cache_dict_name
@@ -657,7 +757,7 @@ class keyed_memoize_on_first_arg:  # noqa: N801
                 return getattr(obj, cache_dict_name)[cache_key]
             except AttributeError:
                 result = function(obj, *args, **kwargs)
-                setattr(obj, cache_dict_name, {cache_key: result})
+                object.__setattr__(obj, cache_dict_name, {cache_key: result})
                 return result
             except KeyError:
                 result = function(obj, *args, **kwargs)
@@ -665,7 +765,7 @@ class keyed_memoize_on_first_arg:  # noqa: N801
                 return result
 
         def clear_cache(obj):
-            delattr(obj, cache_dict_name)
+            object.__delattr__(obj, cache_dict_name)
 
         from functools import update_wrapper
         new_wrapper = update_wrapper(wrapper, function)
@@ -681,9 +781,14 @@ class keyed_memoize_method(keyed_memoize_on_first_arg):  # noqa: N801
         which computes and returns the cache key.
 
     .. versionadded :: 2020.3
+
+    .. versionchanged:: 2021.2
+
+        Can memoize methods on classes that do not allow setting attributes
+        (e.g. by overwritting ``__setattr__``).
     """
     def _default_cache_dict_name(self, function):
-        return intern("_memoize_dic_" + function.__name__)
+        return intern(f"_memoize_dic_{function.__name__}")
 
 
 def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
@@ -708,7 +813,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     uncached_kwargs = list(uncached_kwargs)
 
     def parametrized_decorator(method):
-        cache_dict_name = intern("_memoize_dic_"+method.__name__)
+        cache_dict_name = intern(f"_memoize_dic_{method.__name__}")
 
         def wrapper(self, *args, **kwargs):
             cache_args = list(args)
@@ -725,7 +830,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
                     cache_kwargs.pop(name, None)
 
                 key = (
-                        (_HasKwargs, frozenset(six.iteritems(cache_kwargs)))
+                        (_HasKwargs, frozenset(cache_kwargs.items()))
                         + cache_args)
             else:
                 key = cache_args
@@ -734,7 +839,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
                 return getattr(self, cache_dict_name)[key]
             except AttributeError:
                 result = method(self, *args, **kwargs)
-                setattr(self, cache_dict_name, {key: result})
+                object.__setattr__(self, cache_dict_name, {key: result})
                 return result
             except KeyError:
                 result = method(self, *args, **kwargs)
@@ -742,7 +847,7 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
                 return result
 
         def clear_cache(self):
-            delattr(self, cache_dict_name)
+            object.__delattr__(self, cache_dict_name)
 
         if sys.version_info >= (2, 5):
             from functools import update_wrapper
@@ -767,7 +872,6 @@ def memoize_method_nested(inner):
             "Use @memoize_in(self, 'identifier') instead", DeprecationWarning,
             stacklevel=2)
 
-    from functools import wraps
     cache_dict_name = intern("_memoize_inner_dic_%s_%s_%d"
             % (inner.__name__, inner.__code__.co_filename,
                 inner.__code__.co_firstlineno))
@@ -796,7 +900,7 @@ def memoize_method_nested(inner):
     return new_inner
 
 
-class memoize_in(object):  # noqa
+class memoize_in:  # noqa
     """Adds a cache to a function nested inside a method. The cache is attached
     to *container*.
 
@@ -816,8 +920,6 @@ class memoize_in(object):  # noqa
         self.cache_dict = memoize_in_dict.setdefault(identifier, {})
 
     def __call__(self, inner):
-        from functools import wraps
-
         @wraps(inner)
         def new_inner(*args):
             try:
@@ -866,7 +968,7 @@ def monkeypatch_class(_name, bases, namespace):
 
     assert len(bases) == 1, "Exactly one base class required"
     base = bases[0]
-    for name, value in six.iteritems(namespace):
+    for name, value in namespace.items():
         if name != "__metaclass__":
             setattr(base, name, value)
     return base
@@ -914,8 +1016,7 @@ def flatten(iterable):
     Example: Turn [[a,b,c],[d,e,f]] into [a,b,c,d,e,f].
     """
     for sublist in iterable:
-        for j in sublist:
-            yield j
+        yield from sublist
 
 
 def general_sum(sequence):
@@ -973,18 +1074,14 @@ def partition2(iterable):
     return part_true, part_false
 
 
-def product(iterable):
+def product(iterable: Iterable[Any]) -> Any:
     from operator import mul
     return reduce(mul, iterable, 1)
 
 
-all = six.moves.builtins.all  # pylint: disable=redefined-builtin
-any = six.moves.builtins.any  # pylint: disable=redefined-builtin
-
-
 def reverse_dictionary(the_dict):
     result = {}
-    for key, value in six.iteritems(the_dict):
+    for key, value in the_dict.items():
         if value in result:
             raise RuntimeError(
                     "non-reversible mapping, duplicate key '%s'" % value)
@@ -1415,8 +1512,8 @@ def get_read_from_map_from_permutation(original, permuted):
             "removed in 2019", DeprecationWarning, stacklevel=2)
 
     assert len(original) == len(permuted)
-    where_in_original = dict(
-            (original[i], i) for i in range(len(original)))
+    where_in_original = {
+            original[i]: i for i in range(len(original))}
     assert len(where_in_original) == len(original)
     return tuple(where_in_original[pi] for pi in permuted)
 
@@ -1445,8 +1542,8 @@ def get_write_to_map_from_permutation(original, permuted):
 
     assert len(original) == len(permuted)
 
-    where_in_permuted = dict(
-            (permuted[i], i) for i in range(len(permuted)))
+    where_in_permuted = {
+            permuted[i]: i for i in range(len(permuted))}
 
     assert len(where_in_permuted) == len(permuted)
     return tuple(where_in_permuted[oi] for oi in original)
@@ -1454,23 +1551,6 @@ def get_write_to_map_from_permutation(original, permuted):
 # }}}
 
 
-# {{{ code maintenance
-
-class MovedFunctionDeprecationWrapper:
-    def __init__(self, f):
-        self.f = f
-
-    def __call__(self, *args, **kwargs):
-        from warnings import warn
-        warn("This function is deprecated. Use %s.%s instead." % (
-            self.f.__module__, self.f.__name__),
-            DeprecationWarning, stacklevel=2)
-
-        return self.f(*args, **kwargs)
-
-# }}}
-
-
 # {{{ graph algorithms
 
 from pytools.graph import a_star as a_star_moved
@@ -1486,30 +1566,138 @@ a_star = MovedFunctionDeprecationWrapper(a_star_moved)
 class Table:
     """An ASCII table generator.
 
+    :arg alignments: List of alignments of each column ('l', 'c', or 'r',
+        for left, center, and right alignment, respectively). Columns which
+        have no alignment specifier will use the last specified alignment. For
+        example, with `alignments=['l', 'r']`, the third and all following
+        columns will use 'r' alignment.
+
     .. automethod:: add_row
     .. automethod:: __str__
     .. automethod:: latex
+    .. automethod:: github_markdown
     """
 
-    def __init__(self):
+    def __init__(self, alignments=None):
         self.rows = []
+        if alignments is not None:
+            self.alignments = alignments
+        else:
+            self.alignments = ["l"]
 
     def add_row(self, row):
         self.rows.append([str(i) for i in row])
 
     def __str__(self):
+        """
+        Returns a string representation of the table.
+
+        .. doctest ::
+
+            >>> tbl = Table(alignments=['l', 'r', 'l'])
+            >>> tbl.add_row([1, '|'])
+            >>> tbl.add_row([10, '20||'])
+            >>> print(tbl)
+            1  |    |
+            ---+------
+            10 | 20||
+
+        """
+
         columns = len(self.rows[0])
         col_widths = [max(len(row[i]) for row in self.rows)
                       for i in range(columns)]
 
-        lines = [" | ".join([cell.ljust(col_width)
-            for cell, col_width in zip(row, col_widths)])
+        alignments = self.alignments
+        # If not all alignments were specified, extend alignments with the
+        # last alignment specified:
+        alignments += self.alignments[-1] * (columns - len(self.alignments))
+
+        lines = [" | ".join([
+            cell.center(col_width) if align == "c"
+            else cell.ljust(col_width) if align == "l"
+            else cell.rjust(col_width)
+            for cell, col_width, align in zip(row, col_widths, alignments)])
             for row in self.rows]
         lines[1:1] = ["+".join("-" * (col_width + 1 + (i > 0))
             for i, col_width in enumerate(col_widths))]
 
         return "\n".join(lines)
 
+    def github_markdown(self):
+        r"""Returns a string representation of the table formatted as
+        `GitHub-Flavored Markdown.
+        <https://docs.github.com/en/github/writing-on-github/organizing-information-with-tables>`__
+
+        .. doctest ::
+
+            >>> tbl = Table(alignments=['l', 'r', 'l'])
+            >>> tbl.add_row([1, '|'])
+            >>> tbl.add_row([10, '20||'])
+            >>> print(tbl.github_markdown())
+            1  |     \|
+            :--|-------:
+            10 | 20\|\|
+
+        """  # noqa: W605
+        # Pipe symbols ('|') must be replaced
+        rows = [[w.replace("|", "\\|") for w in r] for r in self.rows]
+
+        columns = len(rows[0])
+        col_widths = [max(len(row[i]) for row in rows)
+                      for i in range(columns)]
+
+        alignments = self.alignments
+        # If not all alignments were specified, extend alignments with the
+        # last alignment specified:
+        alignments += self.alignments[-1] * (columns - len(self.alignments))
+
+        lines = [" | ".join([
+            cell.center(col_width) if align == "c"
+            else cell.ljust(col_width) if align == "l"
+            else cell.rjust(col_width)
+            for cell, col_width, align in zip(row, col_widths, alignments)])
+            for row in rows]
+        lines[1:1] = ["|".join(
+            ":" + "-" * (col_width - 1 + (i > 0)) + ":" if align == "c"
+            else ":" + "-" * (col_width + (i > 0)) if align == "l"
+            else "-" * (col_width + (i > 0)) + ":"
+            for i, (col_width, align) in enumerate(zip(col_widths, alignments)))]
+
+        return "\n".join(lines)
+
+    def csv(self, dialect="excel", csv_kwargs=None):
+        """Returns a string containing a CSV representation of the table.
+
+        :arg dialect: String passed to :func:`csv.writer`.
+        :arg csv_kwargs: Dict of arguments passed to :func:`csv.writer`.
+
+        .. doctest ::
+
+            >>> tbl = Table()
+            >>> tbl.add_row([1, ","])
+            >>> tbl.add_row([10, 20])
+            >>> print(tbl.csv())
+            1,","
+            10,20
+        """
+
+        import csv
+        import io
+
+        if csv_kwargs is None:
+            csv_kwargs = {}
+
+        # Default is "\r\n"
+        if "lineterminator" not in csv_kwargs:
+            csv_kwargs["lineterminator"] = "\n"
+
+        output = io.StringIO()
+        writer = csv.writer(output, dialect, **csv_kwargs)
+        writer.writerows(self.rows)
+
+        return output.getvalue().rstrip(csv_kwargs["lineterminator"])
+
     def latex(self, skip_lines=0, hline_after=None):
         if hline_after is None:
             hline_after = []
@@ -1560,9 +1748,9 @@ def string_histogram(  # pylint: disable=too-many-arguments,too-many-locals
             full = int(floor(scaled))
             eighths = int(ceil((scaled-full)*8))
             if eighths:
-                return full*six.unichr(0x2588) + six.unichr(0x2588+(8-eighths))
+                return full*chr(0x2588) + chr(0x2588+(8-eighths))
             else:
-                return full*six.unichr(0x2588)
+                return full*chr(0x2588)
     else:
         def format_bar(cnt):
             return int(ceil(cnt*width/max_count))*"#"
@@ -1587,13 +1775,13 @@ def word_wrap(text, width, wrap_using="\n"):
     breaks are posix newlines (``\n``).
     """
     space_or_break = [" ", wrap_using]
-    return reduce(lambda line, word: '%s%s%s' %
+    return reduce(lambda line, word: "%s%s%s" %
             (line,
-                space_or_break[(len(line)-line.rfind('\n')-1
-                    + len(word.split('\n', 1)[0])
+                space_or_break[(len(line)-line.rfind("\n")-1
+                    + len(word.split("\n", 1)[0])
                     >= width)],
                     word),
-            text.split(' ')
+            text.split(" ")
             )
 
 # }}}
@@ -1604,12 +1792,12 @@ def word_wrap(text, width, wrap_using="\n"):
 def _exec_arg(arg, execenv):
     import os
     if os.access(arg, os.F_OK):
-        exec(compile(open(arg, "r"), arg, 'exec'), execenv)
+        exec(compile(open(arg), arg, "exec"), execenv)
     else:
-        exec(compile(arg, "<command line>", 'exec'), execenv)
+        exec(compile(arg, "<command line>", "exec"), execenv)
 
 
-class CPyUserInterface(object):
+class CPyUserInterface:
     class Parameters(Record):
         pass
 
@@ -1633,14 +1821,14 @@ class CPyUserInterface(object):
         print()
         print("The following variables are recognized:")
         for v in sorted(self.variables):
-            print("  %s = %s" % (v, self.variables[v]))
+            print("  {} = {}".format(v, self.variables[v]))
             if v in self.doc:
                 print("    %s" % self.doc[v])
 
         print()
         print("The following constants are supplied:")
         for c in sorted(self.constants):
-            print("  %s = %s" % (c, self.constants[c]))
+            print("  {} = {}".format(c, self.constants[c]))
             if c in self.doc:
                 print("    %s" % self.doc[c])
 
@@ -1673,7 +1861,7 @@ class CPyUserInterface(object):
                         "(user variables must start with 'user_' or '_')"
                         % added_key)
 
-        result = self.Parameters(dict((key, execenv[key]) for key in self.variables))
+        result = self.Parameters({key: execenv[key] for key in self.variables})
         self.validate(result)
         return result
 
@@ -1685,7 +1873,7 @@ class CPyUserInterface(object):
 
 # {{{ debugging
 
-class StderrToStdout(object):
+class StderrToStdout:
     def __enter__(self):
         # pylint: disable=attribute-defined-outside-init
         self.stderr_backup = sys.stderr
@@ -1715,17 +1903,17 @@ def typedump(val, max_seq=5, special_handlers=None):
         if isinstance(val, dict):
             return "{%s}" % (
                     ", ".join(
-                        "%r: %s" % (str(k), typedump(v))
-                        for k, v in six.iteritems(val)))
+                        "{!r}: {}".format(str(k), typedump(v))
+                        for k, v in val.items()))
 
         try:
             if len(val) > max_seq:
-                return "%s(%s,...)" % (
+                return "{}({},...)".format(
                         type(val).__name__,
                         ",".join(typedump(x, max_seq, special_handlers)
                             for x in val[:max_seq]))
             else:
-                return "%s(%s)" % (
+                return "{}({})".format(
                         type(val).__name__,
                         ",".join(typedump(x, max_seq, special_handlers)
                             for x in val))
@@ -1755,7 +1943,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
         input("Edit %s at %s now, then hit [Enter]:"
                 % (descr, full_name))
 
-    inf = open(full_name, "r")
+    inf = open(full_name)
     result = inf.read()
     inf.close()
 
@@ -1766,7 +1954,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
 
 # {{{ progress bars
 
-class ProgressBar(object):  # pylint: disable=too-many-instance-attributes
+class ProgressBar:  # pylint: disable=too-many-instance-attributes
     """
     .. automethod:: draw
     .. automethod:: progress
@@ -1814,7 +2002,7 @@ class ProgressBar(object):  # pylint: disable=too-many-instance-attributes
             else:
                 eta_str = "?"
 
-            sys.stderr.write("%-20s [%s] ETA %s\r" % (
+            sys.stderr.write("{:<20} [{}] ETA {}\r".format(
                 self.description,
                 squares*"#"+(self.length-squares)*" ",
                 eta_str))
@@ -1847,7 +2035,7 @@ class ProgressBar(object):  # pylint: disable=too-many-instance-attributes
 def assert_not_a_file(name):
     import os
     if os.access(name, os.F_OK):
-        raise IOError("file `%s' already exists" % name)
+        raise OSError("file `%s' already exists" % name)
 
 
 def add_python_path_relative_to_script(rel_path):
@@ -1937,7 +2125,7 @@ generate_unique_possibilities = MovedFunctionDeprecationWrapper(
         generate_unique_names)
 
 
-class UniqueNameGenerator(object):
+class UniqueNameGenerator:
     """
     .. automethod:: is_name_conflicting
     .. automethod:: add_name
@@ -2002,7 +2190,7 @@ class UniqueNameGenerator(object):
 
 # {{{ recursion limit
 
-class MinRecursionLimit(object):
+class MinRecursionLimit:
     def __init__(self, min_rec_limit):
         self.min_rec_limit = min_rec_limit
 
@@ -2041,8 +2229,13 @@ def download_from_web_if_not_present(url, local_name=None):
         local_name = basename(url)
 
     if not exists(local_name):
-        from six.moves.urllib.request import urlopen
-        with urlopen(url) as inf:
+        from pytools.version import VERSION_TEXT
+        from urllib.request import Request, urlopen
+        req = Request(url, headers={
+            "User-Agent": f"pytools/{VERSION_TEXT}"
+            })
+
+        with urlopen(req) as inf:
             contents = inf.read()
 
             with open(local_name, "wb") as outf:
@@ -2068,14 +2261,14 @@ def find_git_revision(tree_root):  # pylint: disable=too-many-locals
     # https://github.com/numpy/numpy/blob/055ce3e90b50b5f9ef8cf1b8641c42e391f10735/setup.py#L70-L92
     import os
     env = {}
-    for k in ['SYSTEMROOT', 'PATH', 'HOME']:
+    for k in ["SYSTEMROOT", "PATH", "HOME"]:
         v = os.environ.get(k)
         if v is not None:
             env[k] = v
     # LANGUAGE is used on win32
-    env['LANGUAGE'] = 'C'
-    env['LANG'] = 'C'
-    env['LC_ALL'] = 'C'
+    env["LANGUAGE"] = "C"
+    env["LANG"] = "C"
+    env["LC_ALL"] = "C"
 
     from subprocess import Popen, PIPE, STDOUT
     p = Popen(["git", "rev-parse", "HEAD"], shell=False,
@@ -2083,8 +2276,7 @@ def find_git_revision(tree_root):  # pylint: disable=too-many-locals
               cwd=tree_root, env=env)
     (git_rev, _) = p.communicate()
 
-    if sys.version_info >= (3,):
-        git_rev = git_rev.decode()
+    git_rev = git_rev.decode()
 
     git_rev = git_rev.rstrip()
 
@@ -2132,7 +2324,7 @@ def reshaped_view(a, newshape):
 SUPPORTS_PROCESS_TIME = (sys.version_info >= (3, 3))
 
 
-class ProcessTimer(object):
+class ProcessTimer:
     """Measures elapsed wall time and process time.
 
     .. automethod:: __enter__
@@ -2183,7 +2375,7 @@ class ProcessTimer(object):
 
 # {{{ log utilities
 
-class ProcessLogger(object):  # pylint: disable=too-many-instance-attributes
+class ProcessLogger:  # pylint: disable=too-many-instance-attributes
     """Logs the completion time of a (presumably) lengthy process to :mod:`logging`.
     Only uses a high log level if the process took perceptible time.
 
@@ -2296,7 +2488,7 @@ class DebugProcessLogger(ProcessLogger):
     default_noisy_level = logging.DEBUG
 
 
-class log_process(object):  # noqa: N801
+class log_process:  # noqa: N801
     """A decorator that uses :class:`ProcessLogger` to log data about calls
     to the wrapped function.
     """
@@ -2373,6 +2565,107 @@ def natsorted(iterable, key=None, reverse=False):
 # }}}
 
 
+# {{{ resolve_name
+
+# https://github.com/python/cpython/commit/1ed61617a4a6632905ad6a0b440cd2cafb8b6414
+
+_DOTTED_WORDS = r"[a-z_]\w*(\.[a-z_]\w*)*"
+_NAME_PATTERN = re.compile(f"^({_DOTTED_WORDS})(:({_DOTTED_WORDS})?)?$", re.I)
+del _DOTTED_WORDS
+
+
+def resolve_name(name):
+    """A backport of :func:`pkgutil.resolve_name` (added in Python 3.9).
+
+    .. versionadded:: 2021.1.2
+    """
+    # Delete the tail of the function and deprecate this once we require Python 3.9.
+    if sys.version_info >= (3, 9):
+        # use the official version
+        import pkgutil
+        return pkgutil.resolve_name(name)  # pylint: disable=no-member
+
+    import importlib
+
+    m = _NAME_PATTERN.match(name)
+    if not m:
+        raise ValueError(f"invalid format: {name!r}")
+    groups = m.groups()
+    if groups[2]:
+        # there is a colon - a one-step import is all that's needed
+        mod = importlib.import_module(groups[0])
+        parts = groups[3].split(".") if groups[3] else []
+    else:
+        # no colon - have to iterate to find the package boundary
+        parts = name.split(".")
+        modname = parts.pop(0)
+        # first part *must* be a module/package.
+        mod = importlib.import_module(modname)
+        while parts:
+            p = parts[0]
+            s = f"{modname}.{p}"
+            try:
+                mod = importlib.import_module(s)
+                parts.pop(0)
+                modname = s
+            except ImportError:
+                break
+    # if we reach this point, mod is the module, already imported, and
+    # parts is the list of parts in the object hierarchy to be traversed, or
+    # an empty list if just the module is wanted.
+    result = mod
+    for p in parts:
+        result = getattr(result, p)
+    return result
+
+# }}}
+
+
+# {{{ unordered_hash
+
+def unordered_hash(hash_instance, iterable, hash_constructor=None):
+    """Using a hash algorithm given by the parameter-less constructor
+    *hash_constructor*, return a hash object whose internal state
+    depends on the entries of *iterable*, but not their order. If *hash*
+    is the instance returned by evaluating ``hash_constructor()``, then
+    the each entry *i* of the iterable must permit ``hash.upate(i)`` to
+    succeed. An example of *hash_constructor* is ``hashlib.sha256``
+    from :mod:`hashlib`.  ``hash.digest_size`` must also be defined.
+    If *hash_constructor* is not provided, ``hash_instance.name`` is
+    used to deduce it.
+
+    :returns: the updated *hash_instance*.
+
+    .. warning::
+
+        The construction used in this function is likely not cryptographically
+        secure. Do not use this function in a security-relevant context.
+
+    .. versionadded:: 2021.2
+    """
+
+    if hash_constructor is None:
+        from functools import partial
+        import hashlib
+        hash_constructor = partial(hashlib.new, hash_instance.name)
+
+    h_int = 0
+    for i in iterable:
+        h_i = hash_constructor()
+        h_i.update(i)
+        # Using sys.byteorder (for efficiency) here technically makes the
+        # hash system-dependent (which it should not be), however the
+        # effect of this is undone by the to_bytes conversion below, while
+        # left invariant by the intervening XOR operations (which do not
+        # mix adjacent bits).
+        h_int = h_int ^ int.from_bytes(h_i.digest(), sys.byteorder)
+
+    hash_instance.update(h_int.to_bytes(hash_instance.digest_size, sys.byteorder))
+    return hash_instance
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
